from __future__ import annotations

import asyncio
import collections.abc
import json
import logging
import warnings
from contextlib import asynccontextmanager
from dataclasses import dataclass, field, InitVar
from datetime import datetime

import pytest
import pytest_asyncio

from asyncio_multisubscriber_queue import MultisubscriberQueue
from typing import Dict, TYPE_CHECKING


if TYPE_CHECKING:
    from _pytest.fixtures import SubRequest
    from quart.typing import TestClientProtocol
    from typing import Any, AsyncGenerator, Iterator, List, Optional


logger = logging.getLogger(__name__)


def pytest_addoption(parser):
    parser.addini("quart_events_path", "url path for quart-events blueprint")
    parser.addini("quart_events_namespace", "optional namespace for quart-events")


def ignore_cancelled_error(func):
    """
    Decorator for functions that can be cancelled
    without requiring any cleanup
    """

    async def wrapper(*args, **kwargs):
        try:
            await func(*args, **kwargs)
        except asyncio.CancelledError:
            pass

    return wrapper


@pytest_asyncio.fixture(scope="session")
async def quart_events_catcher(
    app_test_client: TestClientProtocol, request: SubRequest
):
    def _getini(name, default=None):
        """
        getini returns an empty string instead of None;
        this helper fixes that
        """
        _val = request.config.getini(name)
        return _val if len(_val) > 0 else default

    """ catch events from quart-events as they are generated in the background """
    _catcher = EventsCatcher(
        app_test_client=app_test_client,
        blueprint_path=_getini("quart_events_path", default="/events"),
        namespace=_getini("quart_events_namespace", default=None),
    )

    async with _catcher:
        yield _catcher


@dataclass(frozen=True)
class Event:
    name: str = field(init=False)
    payload: InitVar[str]
    date: datetime = field(init=False)
    data: Dict = field(init=False)
    seconds_since_last: float = field(init=False)
    last: Event = field(repr=False)

    def __post_init__(self, payload: str):
        object.__setattr__(self, "date", datetime.utcnow())

        try:
            object.__setattr__(self, "data", json.loads(payload))
        except Exception as e:
            logger.error(f"could not decode json for event payload: {payload}")
            raise e

        object.__setattr__(self, "name", self._name())
        object.__setattr__(self, "seconds_since_last", self._seconds_since_last())

    def _name(self) -> Optional[str]:
        if isinstance(self.data, Dict) and "event" in self.data:
            return self.data.pop("event")
        else:
            logger.debug(f'event data does not contain an "event" key: {self.data}')
            return None

    def _seconds_since_last(self) -> float:
        if self.last:
            return self.seconds_since(self.last)
        else:
            return 0.0

    def seconds_since(self, event: Event) -> float:
        assert self is not event
        return (self.date - event.date).total_seconds()

    def get(self, key, default=None) -> Any:
        return self.data.get(key, default)


class EventsCatcher(MultisubscriberQueue):
    def __init__(
        self,
        app_test_client: TestClientProtocol,
        blueprint_path: Optional[str],
        namespace: Optional[str] = None,
    ):
        super().__init__()
        self.app_test_client = app_test_client
        self.blueprint_path = blueprint_path
        self.namespace = namespace
        self._task: Optional[asyncio.Task] = None
        self._ready: asyncio.Event = asyncio.Event()

    async def __aenter__(self):
        self._task = asyncio.create_task(self.run())
        return self

    async def __aexit__(self, *args, **kwargs):
        self._task.cancel()
        await self._task

    def __del__(self):
        if self._task and not self._task.done():
            warnings.warn(f"task for {type(self).__name__} is still running")

    @ignore_cancelled_error
    async def run(self):
        _auth_url = f"{self.blueprint_path}/auth"
        logger.debug(f"authorizing events session via {_auth_url}")
        r = await self.app_test_client.get(_auth_url)

        if r.status_code != 200:
            _payload = await r.get_data()
            logger.error(f"auth request failed [path={_auth_url}]")
            logger.error(f"auth request payload: {_payload}")
            raise RuntimeError("auth request payload")

        data = await r.get_json()
        assert data["authorized"] is True

        url = f"{self.blueprint_path}/ws"
        if self.namespace:
            url = f"{url}/{self.namespace}"

        logger.debug(f"subscribing to events via {self.blueprint_path}/auth")
        _last: Optional[Event] = None
        async with self.app_test_client.websocket(url) as ws:
            _event = await ws.receive()
            _event = json.loads(_event)
            assert _event.get("event") == "_open"

            self._ready.set()

            while True:
                _data = await ws.receive()
                _event = Event(_data, _last)
                if _event.name == "_token_expire":
                    break
                else:
                    await self.put(_event)
                    _last = _event

    def events(self, expected, timeout=5, namespace=None):
        return CaughtEvents(
            catcher=self, expected=expected, timeout=timeout, namespace=namespace
        )


class CaughtEvents:
    def __init__(
        self,
        catcher: EventsCatcher,
        expected: int,
        namespace: Optional[str] = None,
        timeout: int = 5,
    ):
        self.catcher = catcher
        self.expected = expected
        self.namespace = namespace
        self._timout = timeout
        self._task: Optional[asyncio.Task] = None
        self._events: List[Event] = list()

    def __repr__(self) -> str:
        return f"{type(self).__name__} object events={self.event_names()}"

    def __len__(self) -> int:
        return len(self._events)

    def __iter__(self) -> Iterator:
        yield from iter(self._events)

    async def __aenter__(self):
        await asyncio.wait_for(self.catcher._ready.wait(), timeout=5)
        self._task = asyncio.create_task(self.run())
        return self

    async def __aexit__(self, *args, **kwargs):
        try:
            await asyncio.wait_for(self._task, self._timout)
        except asyncio.TimeoutError:
            pass
        await self._task

    def __del__(self):
        if self._task and not self._task.done():
            warnings.warn(f"task for {type(self).__name__} is still running")

    @ignore_cancelled_error
    async def run(self):
        self._events = list()
        async for _event in self.catcher.subscribe():
            if _event.get("event") == "_open" or (
                self.namespace
                and (
                    _event.name is None
                    or _event.name == "_open"
                    or not _event.name.startswith(self.namespace)
                )
            ):
                continue
            self._events.append(_event)
            if len(self._events) >= self.expected:
                break

    def event_names(self) -> List[str]:
        return [event.name for event in self._events if isinstance(event.name, str)]

    def assert_events(self, event_list: List[str]) -> None:
        assert event_list == self.event_names()



from __future__ import annotations

import asyncio
import logging
import functools
import json
from copy import copy
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional
from uuid import UUID, uuid4

from asyncio_multisubscriber_queue import MultisubscriberQueue
from quart import (
    Blueprint,
    make_response,
    jsonify,
    Quart,
    request,
    Response,
    session,
    websocket,
)
from werkzeug.datastructures import Headers

from .errors import EventBrokerError, EventBrokerAuthError


logger = logging.getLogger(__name__)


KeepAlive = object()


@dataclass
class Token:
    value: UUID
    date: datetime

    def __str__(self):
        return str(self.value)

    @staticmethod
    def new():
        return Token(value=uuid4(), date=datetime.utcnow())


class NullToken(Token):
    def __init__(self):
        super().__init__(value=None, date=None)

    def __bool__(self) -> bool:
        return False


class EventBroker(MultisubscriberQueue):
    def __init__(
        self,
        app: Quart,
        url_prefix: str = "/events",
        keepalive: int = 30,
        auth: bool = True,
        token_expire_seconds: int = 3600,
        encoding: str = "utf-8",
    ):
        """
        The constructor for EventBroker class

        Parameters:
            app (quart.Quart): Quart app
            url_prefix (str): prefix for the blueprint
            keepalive (int): how often to send a "keepalive" event when no new
                events are being generated
            auth (bool): enable/disable session validation
            encoding (str): character encoding to use

        """
        if auth is True and app.config.get("SECRET_KEY") is None:
            raise RuntimeError(
                "session support is required for EventBroker authentication; please specify a SECRET_KEY"
            )

        self.keepalive = keepalive
        self.encoding = encoding
        self._auth_enabled: bool = auth
        self._token_expire_seconds = token_expire_seconds
        self._auth_callbacks: List[Callable] = list()
        self._verify_callbacks: List[Callable] = list()
        self._send_callbacks: List[Callable] = list()
        self._tokens: Dict[str, Token] = dict()
        super().__init__()

        if app:
            self.init_app(app, url_prefix)

    def init_app(self, app: Quart, url_prefix: str) -> None:
        """
        Register the blueprint with the application

        """
        app.extensions["events"] = self
        app.register_blueprint(self.create_blueprint(), url_prefix=url_prefix)

    def auth(self, callable_: Callable) -> None:
        self._auth_callbacks.append(callable_)

    def verify(self, callable_: Callable) -> None:
        self._verify_callbacks.append(callable_)

    def send(self, callable_: Callable) -> None:
        self._send_callbacks.append(callable_)

    @staticmethod
    async def _execute_callbacks(callbacks: List[Callable], *args) -> None:
        for _callable in callbacks:
            if asyncio.iscoroutinefunction(_callable):
                await _callable(*args)
            else:
                _callable(*args)

    def _get_token_from_session(self) -> Token:
        _token = session.get("quart_events_token")
        if _token:
            return Token(value=_token["value"], date=_token["date"])
        else:
            return NullToken()

    def _token_is_expired(self, token: Token) -> bool:
        if type(token) is NullToken:
            return True

        _expire_time = datetime.utcnow() - timedelta(seconds=self._token_expire_seconds)
        return token.date < _expire_time

    def clear_expired_tokens(self):
        for _key in list(self._tokens.keys()):
            if self._token_is_expired(self._tokens[_key]):
                del self._tokens[_key]

    async def authorize_websocket(self):
        # gc expired tokens
        self.clear_expired_tokens()

        try:
            await self._execute_callbacks(self._auth_callbacks)

            _token = self._get_token_from_session()
            if _token is type(NullToken) or self._token_is_expired(_token):
                _token = Token.new()
                session["quart_events_token"] = asdict(_token)
                self._tokens[str(_token.value)] = _token
        except Exception as e:
            if "quart_events_token" in session:
                del session["quart_events_token"]
            raise

    def verify_auth_token(self) -> Token:
        _token = self._get_token_from_session()
        if self._token_is_expired(_token):
            raise EventBrokerAuthError("token is expired")
        return _token

    async def verify_auth(self) -> Token:
        # verify a token exists in the user's session
        _token = self.verify_auth_token()

        # executes any additional verification callbacks
        await self._execute_callbacks(self._verify_callbacks)

        return _token

    def create_blueprint(self) -> Blueprint:
        """
        Generate the blueprint

        Returns:
            quart.Blueprint

        """
        blueprint = Blueprint("events", __name__)

        @blueprint.route("/auth")
        async def auth() -> Response:
            try:
                if self._auth_enabled:
                    await self.authorize_websocket()
                return jsonify(authorized=True)
            except EventBrokerAuthError as e:
                r = jsonify(authorized=False, error=str(e))
                r.status_code = 401
                return r
            except Exception as e:
                logger.exception(e)
                r = jsonify(authorized=False, error="an unknown error occurred")
                r.status_code = 400
                return r

        @blueprint.websocket("/ws")
        @blueprint.websocket("/ws/<namespace>")
        async def ws(namespace: Optional[str] = None) -> Response:
            if self._auth_enabled:
                try:
                    _token = await self.verify_auth()
                except EventBrokerAuthError as e:
                    await websocket.send_json({"event": "error", "message": str(e)})
                    return jsonify(error=str(e))
                except Exception as e:
                    await websocket.send_json(
                        {"event": "error", "message": "not authorized"}
                    )
                    return jsonify(error="not authorized")

            # initial message
            await websocket.send_json({"event": "_open"})

            # enter subscriber loop
            async for data in self.subscribe():
                try:
                    """
                    KeepAlive:
                        * dummy event send at a regular interval to keep the socket from closing
                    Namespace:
                        * if a namespace is given but the "event" field is not, skip this event.
                        * if a namespace is given but does not match the "event" field, skip this event.
                    """
                    if self._token_is_expired(_token):
                        await websocket.send_json(
                            {"event": "_token_expire", "message": "token is expired"}
                        )
                        break
                    elif data is KeepAlive:
                        await websocket.send_json({"event": "_keepalive"})
                    elif namespace and (
                        data.get("event") is None
                        or not data["event"].startswith(namespace)
                    ):
                        continue
                    else:
                        await self._execute_callbacks(self._send_callbacks, data)
                        await websocket.send_json(data)
                except asyncio.CancelledError:
                    break
                except Exception as e:
                    logger.exception(e)
                    logger.warning("ending subscriber loop")
                    break
            else:
                # final message
                await websocket.send_json({"event": "_close"})

            return jsonify(message="socket has ended")

        return blueprint

    async def put(self, **data: Any) -> None:  # type: ignore
        """
        Put a new data on the event broker

        """
        if "event" not in data:
            data["event"] = None

        await super().put(data)

    async def subscribe(self) -> AsyncGenerator:
        """
        Override subscribe() to add a timeout for the keepalive event

        """
        with self.queue() as q:
            while True:
                try:
                    _value = await asyncio.wait_for(q.get(), self.keepalive)
                except asyncio.TimeoutError:
                    yield KeepAlive
                else:
                    if _value is StopAsyncIteration:
                        break
                    else:
                        yield _value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quart_events']

package_data = \
{'': ['*']}

install_requires = \
['asyncio-multisubscriber-queue', 'quart']

entry_points = \
{'console_scripts': ['pytest = pytest:main'],
 'pytest11': ['quart_events = quart_events.pytest_plugin']}

setup_kwargs = {
    'name': 'quart-events',
    'version': '0.4.3',
    'description': 'aquart extension to facilitate event message brokering',
    'long_description': "#  quart-events\n\n[![PyPI version](https://img.shields.io/pypi/v/quart-events)](https://pypi.org/project/quart-events/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/quart-events)](https://pypi.org/project/quart-events/)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![](https://github.com/smithk86/quart-events/workflows/pytest/badge.svg)](https://github.com/smithk86/quart-events/actions?query=workflow%3Apytest)\n\n## Usage\n\nquart_events.EventBroker loads a blueprint into Quart which allows clients to subscribe to events via a WebSockets. The app can then generate events that can be sent to all subscribed clients in real-time.\n\nPlease see [test/app.py](https://github.com/smithk86/quart-events/blob/main/test/testapp/) for an example app. This app is used when running testing via py.test but can also be run standalone.\n\n## Change Log\n\n### [0.4.2] - 2021-12-23\n\n- Change build system from setuptools to poetry\n\n### [0.4.0] - 2021-11-08\n\n- add type hints and type validation with mypy\n- requires asyncio-multisubscriber-queue 0.3.0\n- pytest plugin to facilitate capturing events while other tests are running; plugin name is *quart_events_catcher*\n- added optional callbacks\n- websocket auth improvements\n    - token is now seemlessly managed using the user's session data\n    - token has an expiration; user is disconnected from the socket upon expiration\n    - a callback is available to further validate user using other criteria (like Flask-Login)\n",
    'author': 'Kyle Smith',
    'author_email': 'smithk86@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/smithk86/quart-events',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
